package app.billboard;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import logic.sequence.io.db.EnsemblConnection;
import logic.sequence.io.db.QueryException;
import logic.sequence.io.db.SqlLib;
import logic.sequence.io.db.sql.entity.ChromosomeMap;
import logic.sequence.io.db.sql.entity.Dnafrag;
import logic.sequence.io.db.sql.entity.Exon;
import logic.sequence.io.db.sql.entity.Gene;
import logic.sequence.io.db.sql.entity.GenomicAlign;
import logic.sequence.io.db.sql.entity.Transcript;
import logic.sequence.io.db.sql.entity.manager.StringManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.PropertyConfigurator;

import app.billboard.homologs.CigarLine;
import app.ensembl.Species;

public class Homologs {
	private static final Log log = LogFactory.getLog(Homologs.class);

	// seq_region_id -> name
	private static Map<Integer, String> homoChromosomes = new HashMap<Integer, String>();
	private static EnsemblConnection c;
	private static List<Gene> homoGens = new LinkedList<Gene>();
	private static BufferedWriter results;
	private static final String hGenesPath = "res/hGenes";
	private static BufferedWriter hGenes;

	private static Map<Integer, String> mouseChromosomes = new HashMap<Integer, String>();

	static {
		PropertyConfigurator.configure("log4j.properties");
	}

	public static void main(String[] args) {
		try {
			c = EnsemblConnection.getConnection();
		} catch (Exception e) {
			log.fatal("Could not initialise connection", e);
			return;
		}
		
		try {
		populateHomoChromosomes();
		log.info("Human chromosomes:");
		log.info(homoChromosomes.toString());

		populateMouseChromosomes();
		log.info("Mouse chromosomes:");
		log.info(mouseChromosomes.toString());
		} catch (Exception e) {
			log.fatal("Strange db problem.", e);
			return;
		}

		Set<String> processedGenes;
		try {
		results = new BufferedWriter(new FileWriter("output/"
				+ System.currentTimeMillis() + "-pairs"));
		File hGenesFile = new File(hGenesPath);
		processedGenes = new HashSet<String>();
		if (hGenesFile.exists()) {
			BufferedReader reader = new BufferedReader(new FileReader(
					hGenesFile));
			String line;
			while ((line = reader.readLine()) != null) {
				processedGenes.add(line.trim());
			}
		}
		hGenes = new BufferedWriter(new FileWriter(hGenesPath));
		} catch (Exception e) {
			log.fatal("Strange filesystem problem.", e);
			return;
		}

		try {
			List<Gene> homoGenesSet = (List<Gene>) c.list(SqlLib.getGenesForSpecies(Species.HOMO_SAPIENS), Gene.getEntityManager());
			for (Gene hGene : homoGenesSet) {
				if (!processedGenes.contains(hGene.stableId))
					homoGens.add(hGene);
			}
		} catch (Exception e) {
			log.fatal("Strange db problem.", e);
			return;
		}
		log.info("Number of human genes: " + homoGens.size());
		long cur = System.currentTimeMillis();
		try {
			for (int k = 0; k < homoGens.size(); k++) {
				long upto = System.currentTimeMillis() - cur;
				Gene gene = homoGens.get(k);
				log.info(k + "/" + homoGens.size() + ". Est: " + upto
						* (homoGens.size() - k - 1) / (k + 1) / 1000 / 60
						+ " minutes");
				String sql = SqlLib.getGeneHomologs(gene.stableId);
				List<String> mouseHomologIds = (List<String>) c.list(sql, new StringManager());

				log.info("Number of homologs for " + gene.stableId + ": "
						+ mouseHomologIds.size());

				if (mouseHomologIds.size() == 0) {
					log.warn("No homologs found for " + gene.stableId);
					continue;
				}

				String sql3 = SqlLib.getTranscriptsForStableId(gene.stableId,
						Species.HOMO_SAPIENS);
				List<Transcript> homoTranscripts = (List<Transcript>) c.list(sql3, Transcript.getEntityManager());
				log.info("Number of transcripts for " + gene.stableId + ": "
						+ homoTranscripts.size());
				if (homoTranscripts.size() == 0) {
					log.warn("No transcripts found for " + gene.stableId);
					continue;
				}
				Map<Integer, Exon> firstExonForHumanTranscript = new HashMap<Integer, Exon>();
				for (Transcript t : homoTranscripts) {
					String firstExon = SqlLib.getFirstExon(t.transcriptId,
							Species.HOMO_SAPIENS);
					Exon exon;
					try {
						exon = (Exon) c.uniqueResult(firstExon, Exon
								.getEntityManager());
					} catch (QueryException e) {
						log.error(e.getMessage());
						continue;
					}
					if (checkRange(exon, firstExonForHumanTranscript.values())) {
						log.info("Transcript in range of another transcript: " + t.stableId);
					} else {
						firstExonForHumanTranscript.put(t.transcriptId, exon);
					}
				}

				for (String homolog : mouseHomologIds) {
					String sql4 = SqlLib.getTranscriptsForStableId(homolog,
							Species.MUS_MUSCULUS);
					List<Transcript> mouseTranscripts = (List<Transcript>) c.list(sql4, Transcript.getEntityManager());
					if (mouseTranscripts.size() == 0) {
						log.warn("No transcripts for " + homolog);
						continue;
					}
					Map<Integer, Exon> firstExonForMouseTranscript = new HashMap<Integer, Exon>();
					for (Transcript mTr : mouseTranscripts) {
						String firstExon = SqlLib.getFirstExon(
								mTr.transcriptId, Species.MUS_MUSCULUS);
						Exon e;
						try {
							e = (Exon) c.uniqueResult(firstExon, Exon
									.getEntityManager());
						} catch (QueryException e1) {
							log.error(e1.getMessage());
							continue;
						}
						if (!checkRange(e, firstExonForMouseTranscript.values())) {
							firstExonForMouseTranscript
									.put(mTr.transcriptId, e);
						}
					}

					for (Transcript hTr : homoTranscripts) {
						Exon hEx = firstExonForHumanTranscript
								.get(hTr.transcriptId);
						if (hEx == null)
							continue;
						List<GenomicAlign> alignRegions;
						try {
							alignRegions = getGlobalAlignmentRegion(
									hEx.seqRegionId, hEx.seqRegionStrand ?
											hEx.seqRegionStart : hEx.seqRegionEnd);
						} catch (QueryException e1) {
							log.warn(e1);
							continue;
						}
						if (alignRegions == null) {
							log.warn("No genomic alignment found");
							continue;
						}
						if (alignRegions.size() > 1) {
							log.warn("More than one region");
							continue;
						}
						GenomicAlign ar = getRegion(hEx, alignRegions);
						if (ar == null) {
							log.warn("No genomic alignment found for position");
							continue;
						}
						GenomicAlign arPair;
						try {
							arPair = (GenomicAlign) c.uniqueResult(SqlLib
									.getPairedAlignedRegion(
											ar.genomic_align_block_id,
											ar.genomic_align_id), GenomicAlign
									.getEntityManager());
						} catch (QueryException e) {
							log.error(e);
							continue;
						}
						int mOffset = computeExonAlignedRegion(hEx, ar, arPair);
						int x = mOffset + arPair.dnafrag_start;
						
						Map<Exon, Transcript> helper = new HashMap<Exon, Transcript>();
						Exon candidate = null;
						Dnafrag mDnafrag = (Dnafrag) c.uniqueResult(SqlLib.getDnafragById(arPair.dnafrag_id), Dnafrag.getEntityManager());
						for (Transcript mTr : mouseTranscripts) {
							Exon mEx = firstExonForMouseTranscript
									.get(mTr.transcriptId);
								if (mEx != null) {
									String mChromosome = mouseChromosomes.get(mEx.seqRegionId);
									if (mChromosome != null && mDnafrag.name.equals(mChromosome)) {
										helper.put(mEx, mTr);
										int mPos = mEx.seqRegionStrand ? mEx.seqRegionStart : mEx.seqRegionEnd;
										candidate = (candidate == null ? mEx
												: Math
												.abs(mPos
														- x) < Math
														.abs((candidate.seqRegionStrand ? candidate.seqRegionStart : candidate.seqRegionEnd)
																- x) ? mEx
																		: candidate);
									}
								}
						}
						if (candidate != null) {
							int difference = Math.abs((candidate.seqRegionStrand ? candidate.seqRegionStart : candidate.seqRegionEnd) - x);
							log.info("Alignment start difference for " + hTr.stableId + " : " + difference);
							String str = difference + "," + gene.stableId + "," + hTr.stableId + "," + helper.get(candidate).stableId;
							if (difference < 10000) {
								results.write(hTr.stableId + "\t"
										+ helper.get(candidate).stableId);
								results.newLine();
							} else {
								log.info("Alignment difference too big for " + hTr.stableId);
							}
						}
					}
				}
				hGenes.write(gene.stableId);
				hGenes.newLine();
			}
		} catch (Exception e) {
			log.fatal("Unexpected error", e);
		} finally {
			try {
				hGenes.flush();
				hGenes.close();
			} catch (IOException e) {
				log.fatal("What the....???!!", e);
			}
			try {
				results.flush();
				results.close();
			} catch (IOException e) {
				log.fatal("What the....???!!", e);
			}
		}
	}

	private static GenomicAlign getRegion(Exon ex,
			List<GenomicAlign> alignRegions) {
		int pos = ex.seqRegionStrand ? ex.seqRegionStart : ex.seqRegionEnd;
		for (GenomicAlign genomicAlign : alignRegions) {
			if (pos < genomicAlign.dnafrag_end || pos >= genomicAlign.dnafrag_start)
				return genomicAlign;
		}
		return null;
	}

	private static int computeExonAlignedRegion(Exon ex, GenomicAlign ar,
			GenomicAlign arPair) {
		int dist = ex.seqRegionStrand ? ex.seqRegionStart - ar.dnafrag_start : ex.seqRegionEnd - ar.dnafrag_start;
		CigarLine mCigar = new CigarLine(arPair.cigar_line, arPair.dnafrag_strand);
		CigarLine hCigar = new CigarLine(ar.cigar_line, ar.dnafrag_strand);
		int hOffset = hCigar.getOffset(dist);
		int mOffset = mCigar.countOffset(hOffset);
		if (arPair.dnafrag_strand)
			return mOffset;
		else
			return arPair.dnafrag_end - arPair.dnafrag_start - mOffset + 1;
	}

	private static List<GenomicAlign> getGlobalAlignmentRegion(int seqRegionId,
			int seqRegionPos) throws SQLException,
			QueryException {
		String regionName;
		if ((regionName = homoChromosomes.get(seqRegionId)) == null) {
			log.error("Not on chromosome (seqRegionId: " + seqRegionId + ")");
			return null;
		}
		Dnafrag d = (Dnafrag) c.uniqueResult(SqlLib.getDnafrag(regionName,
				Species.HOMO_SAPIENS), Dnafrag.getEntityManager());
		String alignedRegionSql = SqlLib.getAlignedRegion(d.dnafragId,
				seqRegionPos);
		log.debug(regionName + ": " + alignedRegionSql);
		List<GenomicAlign> genomicAligns = (List<GenomicAlign>) c.list(alignedRegionSql, GenomicAlign.getEntityManager());
		if (genomicAligns.size() > 1) {
			Collections.sort(genomicAligns, new Comparator<GenomicAlign>() {
				public int compare(GenomicAlign o1, GenomicAlign o2) {
					return o1.dnafrag_start - o2.dnafrag_start;
				}
			});
		}
		if (genomicAligns.size() == 0)
			return null;
		return genomicAligns;
	}

	private static boolean checkRange(Exon exon, Collection<Exon> exons) {
		for (Exon e : exons) {
			if (inRange(e, exon)) {
				return true;
			}
		}
		return false;
	}

	private static boolean inRange(Exon e, Exon exon) {
		if (e.seqRegionId == exon.seqRegionId) {
			return !((e.seqRegionStart < exon.seqRegionStart && e.seqRegionEnd < exon.seqRegionStart) || (e.seqRegionStart > exon.seqRegionEnd && e.seqRegionEnd > exon.seqRegionEnd));
		} else {
			return false;
		}
	}

	private static void populateMouseChromosomes() throws SQLException {
		List<ChromosomeMap> mch = (List<ChromosomeMap>) c.list(SqlLib.mouseChromosomes, ChromosomeMap.getEntityManager());
		for (ChromosomeMap cm : mch) {
			mouseChromosomes.put(cm.id, cm.name);
		}
	}

	// inclusive
/*	@SuppressWarnings("unused")
	private static String getSequence(int chromosomeSegRegion, Species s,
			int first, int last, boolean strand) throws SQLException,
			IOException {
		if (s.equals( Species.MUS_MUSCULUS )) {
			if (mouseChromosomes.get(chromosomeSegRegion) == null) {
				log.error("Not on chromosome (seqRegionId: "
						+ chromosomeSegRegion + ")");
				return null;
			}
		} else {
			if (homoChromosomes.get(chromosomeSegRegion) == null) {
				log.error("Not on chromosome (seqRegionId: "
						+ chromosomeSegRegion + ")");
				return null;
			}
		}
		String str = SqlLib.getAssemblyFragments(chromosomeSegRegion, first,
				last, s);
		List<AssemblyRegion> assemblies = (List<AssemblyRegion>) c.list(str, AssemblyRegion.getEntityManager());
		Collections.sort(assemblies, new Comparator<AssemblyRegion>() {
			public int compare(AssemblyRegion a, AssemblyRegion b) {
				return a.asmStart - a.asmEnd;
			}
		});

		Iterator<AssemblyRegion> it = assemblies.iterator();
		AssemblyRegion ar = null;
		int lastRead = first - 1;
		StringBuffer sb = new StringBuffer();
		do {
			ar = it.next();
			//$ANALYSIS-IGNORE
			int start, end;
			if (ar.asmStart > lastRead) {
				System.err.println("Unexpected gap. " + ar);
				break;
			} else {
				start = lastRead - ar.asmStart + 1 + ar.cmpStart;
			}
			if (ar.asmEnd >= last) {
				end = last - ar.asmStart + ar.cmpStart;
			} else {
				end = ar.cmpEnd;
			}

			lastRead += (end - start + 1);
			String sql = SqlLib.getSequence(ar.cmpId, start, end, s);
			ResultSet seq = c.execute(sql);
			seq.first();
			sb.append(seq.getString(1));
			seq.close();
		} while (it.hasNext() && lastRead < last);
		String sequence = strand ? sb.toString() : reverseComplement(sb
				.toString());
		return sequence;
	}*/

	private static void populateHomoChromosomes() throws SQLException {
		List<ChromosomeMap> list = (List<ChromosomeMap>) c.list(SqlLib.homoChromosomes, ChromosomeMap.getEntityManager());
		for (ChromosomeMap cm : list) {
			homoChromosomes.put(cm.id, cm.name);
		}
	}

	private static String reverseComplement(String seq) {
		StringBuilder sb = new StringBuilder();
		for (int i = seq.length() - 1; i >= 0; i--) {
			char r = reverse(seq.charAt(i));
			sb.append(r);
		}
		return sb.toString();
	}

	private static final char reverse(char r) {
		switch (r) {
		case 'a':
		case 'A':
			return 'T';
		case 't':
		case 'T':
			return 'A';
		case 'g':
		case 'G':
			return 'C';
		case 'c':
		case 'C':
			return 'G';
		default:
			return r;
		}
	}

}
