package app.engine;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import logic.alignment.AlignmentPrepared;
import logic.sequence.DNASequence;
import logic.sequence.impl.BadAlphabet;
import logic.sequence.impl.PreparedSequence;
import logic.sequence.io.EnsembleWSReader;
import logic.tfbs.PatternSet;
import logic.tfbs.impl.NoSuchLetter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import app.Config;
import app.Config.PropName;
import app.ensembl.IdNotRecognizedException;
import app.ensembl.Species;
import app.exception.BillboardException;
import app.intf.Parameters;
import app.test.ParameterParser;

public class BillboardEngine {
	
	protected final Log logger = LogFactory.getLog(getClass());
	
	private Map<String, PatternSet> patternSets;
	private Map<String, Map<String, DNASequence>> backgroundSequences;
	private EnsembleWSReader ensReader;
	private Parameters parameters;
	
	private String propertyFilePath;
	private Properties appProperties;
	
	private EngineInitTools engineInitTools;
	private ResourceLoader resourceLoader;
	
	public void init() throws FileNotFoundException, IOException, BillboardException {
		appProperties = new Properties();
		File propertyFile = new File(resourceLoader.getResourceRealPath(propertyFilePath));
		logger.info("Using property file: " + propertyFile.getAbsolutePath());
		appProperties.load(new FileInputStream(propertyFile));
		parameters = (new ParameterParser(new BufferedReader(new FileReader(new File(appProperties.getProperty(PropName.PARAMETERS_FILE)))), appProperties).parse()).get(0);
		logger.info("Using parameters from property file: " + appProperties);
		patternSets = engineInitTools.getPatterns(appProperties.getProperty(Config.PropName.DEFAULT_TF_DIR));
		backgroundSequences = engineInitTools.getBackgroundSequences(appProperties.getProperty(Config.PropName.BACKGROUND_SEQUENCES_DIR));
	}
	
	public String[] findHomologs(String ensembleId, Species otherOrganism) throws IOException, IdNotRecognizedException {
		return ensReader.getHomologs(ensembleId, otherOrganism);
	}
	
	public String[] listPatternSets() {
		return patternSets.keySet().toArray(new String[0]);
	}
	
	public Species[] listSpecies() {
		return Species.values();
	}
	
	public AlignmentPrepared findAlignment(String onThis, String withThis, String patternSet) throws IOException, BadAlphabet, BillboardException {
		DNASequence onThisSeq = ensReader.getEnsembleSequenceById(onThis, true, 10000, 1000);
		DNASequence withThisSeq = ensReader.getEnsembleSequenceById(withThis, true, 10000, 1000);
		
		if (onThisSeq == null || withThisSeq == null)
			throw new BillboardException("Could not retrieve one of the sequences: " + onThis + ", " + withThis);
		
		PatternSet set = patternSets.get(patternSet);
		if (set == null)
			throw new BillboardException("Could not find a pattern set with name " + patternSet);
		
		PreparedSequence onThisPrep = new PreparedSequence(onThisSeq, set, parameters.getThresholdType());
		PreparedSequence withThisPrep = new PreparedSequence(withThisSeq, set, parameters.getThresholdType());
		
		Species onThisSpecies = Species.getForEnsembleId(onThis);
		Species withThisSpecies = Species.getForEnsembleId(withThis);
		
		if (onThisSpecies == null || withThisSpecies == null)
			throw new BillboardException("Spiecies not supported for either " + onThis + " or " + withThis);
			
		Map<String, DNASequence> bgSeqs = getBackgroundSequences().get(withThisSpecies.getDataSet());
		
		if (bgSeqs == null)
			throw new BillboardException("Could not find background sequences for " + withThisSpecies);
		
		Set<PreparedSequence> preparedBgSeqs = new HashSet<PreparedSequence>();
		for (DNASequence s : bgSeqs.values()) {
			preparedBgSeqs.add(new PreparedSequence(s, set, parameters.getThresholdType()));
		}
		return findAlignment(onThisPrep, withThisPrep, parameters, preparedBgSeqs);
	}
	
	public AlignmentPrepared findAlignment(PreparedSequence onThis, PreparedSequence withThis, Parameters parameters, Set<PreparedSequence> bgSequences) throws NoSuchLetter {
		return new AlignmentPrepared(onThis, withThis, parameters, bgSequences).align();
	}
	
	public Map<String, Map<String, DNASequence>> getBackgroundSequences() {
		return backgroundSequences;
	}
	public void setBackgroundSequences(
			Map<String, Map<String, DNASequence>> backgroundSequences) {
		this.backgroundSequences = backgroundSequences;
	}
	public EnsembleWSReader getEnsReader() {
		return ensReader;
	}
	public void setEnsReader(EnsembleWSReader ensReader) {
		this.ensReader = ensReader;
	}
	public Parameters getParameters() {
		return parameters;
	}
	public void setParameters(Parameters parameters) {
		this.parameters = parameters;
	}
	public Map<String, PatternSet> getPatternSets() {
		return patternSets;
	}
	public void setPatternSets(Map<String, PatternSet> patternSets) {
		this.patternSets = patternSets;
	}
	public String getPropertyFilePath() {
		return propertyFilePath;
	}
	public void setPropertyFilePath(String propertyFilePath) {
		this.propertyFilePath = propertyFilePath;
	}

	public void setEngineInitTools(EngineInitTools engineInitTools) {
		this.engineInitTools = engineInitTools;
	}

	public void setResourceLoader(ResourceLoader resourceLoader) {
		this.resourceLoader = resourceLoader;
	}
	
}
