package app.engine;

import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import logic.sequence.DNASequence;
import logic.sequence.impl.BadAlphabet;
import logic.sequence.io.FastaSequenceReader;
import logic.tfbs.PSSM;
import logic.tfbs.PatternSet;
import logic.tfbs.TFPattern;
import logic.tfbs.TFPatternBuilder;
import logic.tfbs.impl.Background;
import logic.tfbs.impl.PSSMImpl.PSSMIOFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import app.ensembl.IdNotRecognizedException;
import app.ensembl.Species;
import app.exception.ConfigurationException;

public class EngineInitTools {

	protected final Log logger = LogFactory.getLog(getClass());

	private TFPatternBuilder tfPatternBuilder;
	private PSSMIOFactory pssmFactory;
	private ResourceLoader resourceLoader;

	public void setResourceLoader(ResourceLoader resourceLoader) {
		this.resourceLoader = resourceLoader;
	}

	public Map<String, PatternSet> getPatterns(String tfDir) throws ConfigurationException {
		if (tfDir == null)
			throw new NullPointerException("Cannot pass null as TF directory path.");

		Map<String, PatternSet> result = new HashMap<String, PatternSet>();

		File pssmDir = new File(resourceLoader.getResourceRealPath(tfDir));
		logger.info("Using pattern directory: " + pssmDir.getAbsolutePath());
		
		if (!pssmDir.exists())
			throw new ConfigurationException("Directory " + pssmDir.getAbsolutePath() + " does not exist.");
		
		File[] tfDirs = pssmDir.listFiles(new FileFilter() {

			public boolean accept(File file) {
				return file.isDirectory() && !file.getName().startsWith(".");
			}
			
		});

		for (File tfSetDir : tfDirs) {
			File[] pssmFiles = tfSetDir.listFiles(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					if (name.endsWith(".pfm"))
						return true;
					return false;
				}
			});


			List<PSSM> pssms = new LinkedList<PSSM>();

			for (int i = 0; i < pssmFiles.length; i++) {
				try {
					PSSM pssm = pssmFactory.buildSimple(
							new FileReader(pssmFiles[i]), pssmFiles[i].getName().split("\\.")[0]);
					pssms.add(pssm);
				} catch (IOException e) {
					logger.warn("Could not create pssm from file " + pssmFiles[i].getName(), e);
				}
			}

			final List<TFPattern> tfs = new LinkedList<TFPattern>();

			for (final PSSM pssm : pssms) {
				try {
					TFPattern pattern = tfPatternBuilder.makeTFPattern(pssm, Background.getUniformBackground(pssm.getAlphabeth()));
					tfs.add(pattern);
				} catch (Exception e) {
					logger.warn("Could not parse pattern from file " + pssm.getName(), e);
				}
			}

			TFPattern[] patterns = tfs.toArray(new TFPattern[0]);
			Arrays.sort(patterns, new Comparator<TFPattern>() {
				public int compare(TFPattern o1, TFPattern o2) {
					return o1.getName().compareTo(o2.getName());
				}
			});

			PatternSet set = new PatternSet(tfSetDir.getName(), patterns);
			result.put(set.getName(), set);
		}
		return result;
	}

	public Map<String, Map<String, DNASequence>> getBackgroundSequences(String path) throws ConfigurationException {
		if (path == null)
			throw new NullPointerException("Cannot pass null as background sequences directory path.");
		
		Map<String, Map<String, DNASequence>> result = new HashMap<String, Map<String,DNASequence>>();
		
		File file = new File(resourceLoader.getResourceRealPath(path));

		if (!file.exists() || !file.isDirectory()) {
			throw new ConfigurationException("Directory " + path + " does not exist");
		}
		File[] dirs = file.listFiles(new FileFilter() {

			public boolean accept(File file) {
				return file.isDirectory() && !file.getName().startsWith(".");
			}
			
		});
		for (int j = 0; j < dirs.length; j++) {
			File[] files = dirs[j].listFiles(new FileFilter() {

				public boolean accept(File file) {
					return file.isFile();
				}
				
			});
			int read = 0;
			for (int i = 0; i < files.length && read < 100; i++) {
				try {
					for (DNASequence s : FastaSequenceReader.readDNASeqFromFile(files[i], false)) {
						Species species = Species.getForEnsembleId(s.getName());
						Map<String, DNASequence> name = result.get(species.getDataSet());
						if (name == null) {
							name = new HashMap<String, DNASequence>();
							result.put(species.getDataSet(), name);
						}
						if (name.put(s.getName(), s) != null)
							logger.warn("Background sequence " + s.getName() + " registered");
						else
							read++;
					}
					logger.info("Read " + read + " background sequences from " + dirs[j].getName());
				} catch (BadAlphabet e) {
					logger.warn("Sequence over wrong alphabet (probably ambigous nucleotide). File: " + files[i].getName(), e);
				} catch (IOException e) {
					logger.warn("Could not read sequence. File: " + files[i].getName(), e);
				} catch (IdNotRecognizedException e) {
					logger.warn("Could not recognize this sequence's ensembl id", e);
				}
			}
		}
		return result;
	}

	public void setPssmFactory(PSSMIOFactory pssmFactory) {
		this.pssmFactory = pssmFactory;
	}

	public void setTfPatternBuilder(TFPatternBuilder tfPatternBuilder) {
		this.tfPatternBuilder = tfPatternBuilder;
	}
}
