package app.intf;

import java.lang.management.ManagementFactory;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;

import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;

import logic.alignment.AlignmentPrepared;
import logic.alignment.AlignmentPrepared.Footprint;
import logic.sequence.DNASequence;
import logic.sequence.impl.PreparedSequence;
import logic.tfbs.PatternSet;
import logic.threads.OneTimeWorker;
import logic.threads.OneTimeWorker.WorkerObject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import app.BackgroundSequenceManager;
import app.Registry;
import app.SequenceManager;
import app.Helpers.SequencePair;
import app.ensembl.Species;
import app.exception.BillboardException;
import app.exception.ConfigurationException;

public class Experiment {
	private Parameters[] pList;

	protected Log logger = LogFactory.getLog(getClass());

	private Map<String, Set<PreparedSequence>> bgPreps;

	private PatternSet patternSet;

	private List<SequencePair> pairs;

	private BackgroundSequenceManager backgroundSequenceManager;

	private SequenceManager sequenceManager;

	public Experiment(List<Parameters> pList, String patternSetName, List<SequencePair> pairs, SequenceManager sequenceManager, BackgroundSequenceManager backgroundSequenceManager)
			throws ConfigurationException {
		this.pList = pList.toArray(new Parameters[0]);
		this.sequenceManager = sequenceManager;
		this.backgroundSequenceManager = backgroundSequenceManager;
		this.patternSet = Registry.getRegistry().getPatternSet(patternSetName);
		this.pairs = pairs;
		if (patternSet == null) {
			throw new ConfigurationException(
					"Could not find pattern set with name: " + patternSetName);
		}
	}

	public void perform() {
		Registry r = Registry.getRegistry();
		
/*		logger.info("Preparing sequences...");
		if (preps == null)
			prepareSequences();*/

		for (final Parameters p : pList) {
			logger.info("Parameters : " + p);
			OneTimeWorker<AlignmentPrepared> w = new OneTimeWorker<AlignmentPrepared>(p.getThreadNo()) {
				@Override
				public void afterExecute(AlignmentPrepared result) {
					process(result, result.extractClustersDifferently());
				}
				
				@Override
				public void onException(Exception e) {
					logger.error("An exception occured while trying to find an alignment.", e);
				}
			};
			
			for (final SequencePair pair : pairs) {
				w.submit(new Callable<AlignmentPrepared>() {
					public AlignmentPrepared call() throws Exception {
						DNASequence ds1 = sequenceManager.getSequence(pair.first);
						DNASequence ds2 = sequenceManager.getSequence(pair.second);
						if (ds1 == null || ds2 == null) {
							throw new BillboardException("Skipping pair (" + pair.first + ", "
									+ pair.second + ")");
						}
						else {
							PreparedSequence s1 = new PreparedSequence(ds1, Experiment.this.patternSet,
									Experiment.this.pList[0].getThresholdType());
							PreparedSequence s2 = new PreparedSequence(ds2, Experiment.this.patternSet,
									Experiment.this.pList[0].getThresholdType());
							logger.info("Aligning " + s1.getName() + " with " + s2.getName());
							AlignmentPrepared ap;
							ap = new AlignmentPrepared(s1, s2, p, backgroundSequenceManager.getSequences()).align();
							return ap;
						}
					}
				});
			}
			WorkerObject wo = w.execute();
			MBeanServer mbs = ManagementFactory.getPlatformMBeanServer(); 
		    ObjectName name = null;
			try {
				name = new ObjectName("logic.threads:type=WorkerObject");
				mbs.registerMBean(wo, name);
			} catch (MalformedObjectNameException e) {
				e.printStackTrace();
			} catch (NullPointerException e) {
				e.printStackTrace();
			} catch (InstanceAlreadyExistsException e) {
				e.printStackTrace();
			} catch (MBeanRegistrationException e) {
				e.printStackTrace();
			} catch (NotCompliantMBeanException e) {
				e.printStackTrace();
			} 
			wo.run();
			try {
				wo.waitComplete();
			} catch (InterruptedException e) {
				wo.interrupt();
			}
			try {
				mbs.unregisterMBean(name);
			} catch (InstanceNotFoundException e1) {
				e1.printStackTrace();
			} catch (MBeanRegistrationException e1) {
				e1.printStackTrace();
			}
		}
	}

	private Set<PreparedSequence> prepare(DNASequence[] bgSequences) {
		Set<PreparedSequence> result = new HashSet<PreparedSequence>();
		for (DNASequence sequence : bgSequences) {
			result.add(new PreparedSequence(sequence, patternSet, pList[0].getThresholdType()));
		}
		return result;
	}

	/**
	 * This is called once alignment is processed. Should be overriden.
	 * 
	 * @param ap
	 * @param f
	 */
	public void process(AlignmentPrepared ap, List<Footprint> f) {
	}

/*	public class Result {
		public final Parameters p;

		public final List<Footprint> footprints;

		private Result(Parameters parameters, List<Footprint> footprints) {
			this.p = parameters;
			Collections.sort(footprints, new Comparator<Footprint>() {

				public int compare(Footprint o1, Footprint o2) {
					return o1.scorepv < o2.scorepv ? -1 : 1;
				}
				
			});
			this.footprints = footprints;
		}
	}*/

/*	public List<Result> getResults() {
		return results;
	}*/
}
