package app.test;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Callable;

import logic.alignment.AlignmentPrepared;
import logic.alignment.AlignmentPrepared.Footprint;
import logic.tfbs.PSSM;
import logic.tfbs.PatternSet;
import logic.tfbs.TFPattern;
import logic.tfbs.impl.Background;
import logic.tfbs.impl.Threshold;
import logic.threads.OneTimeWorker;
import logic.threads.OneTimeWorker.WorkerObject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.BasicConfigurator;

import app.BackgroundSequenceManager;
import app.Config;
import app.DownloadingSequenceManager;
import app.Helpers;
import app.LocalBackgroundSequenceManager;
import app.LocalSequenceManager;
import app.Registry;
import app.SequenceManager;
import app.Config.PropName;
import app.Helpers.SequencePair;
import app.exception.ConfigurationException;
import app.exception.UserDataException;
import app.intf.Experiment;
import app.intf.Parameters;
import app.test.ArgumentParser.ArgumentParserException;

public class Application {

	private static Log l = LogFactory.getLog(Application.class);

	private static Registry r = Registry.getRegistry();
	
	private static Properties p;
	
	private static String resultFileName = "result.gff";
	
	static {
		BasicConfigurator.configure();
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		initialize();
		//MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
		
		// properties setup
		propertiesSetup(args);

		// processing patterns
		Application.getPatterns(p.getProperty(Config.PropName.DEFAULT_TF_DIR));
		l.info("Got " + r.getPatternSetNames().size() + " pattern set(s).");

		// processing file pairs
		List<SequencePair> pairs = pairsSetup();
		l.info("Got " + pairs.size() + " pairs");
		
		String patternSet = p.getProperty(PropName.TF_SET);
		
		// get sequences
		SequenceManager sequenceManager;
		try {
			sequenceManager = sequenceSetup(pairs);
		} catch (ConfigurationException e2) {
			l.fatal("Error occured while getting sequences. Exiting.", e2);
			return;
		}

		List<Parameters> pList = new LinkedList<Parameters>();
		if (p.getProperty(PropName.PARAMETERS_FILE) == null) {
			l.fatal("Could not find " + PropName.PARAMETERS_FILE + " property. Exiting.");
			return;
		} else {
			try {
				pList = new ParameterParser(new BufferedReader(new FileReader(new File(p.getProperty(PropName.PARAMETERS_FILE)))), p).parse();
			} catch (Exception e) {
				l.fatal("Could not load application parameters.", e);
				return;
			}
		}

		// get background sequences
		BackgroundSequenceManager backgroundSequenceManager;
		try {
			backgroundSequenceManager = backgroundSetup(pList.get(0).getThresholdType(), patternSet);
		} catch (ConfigurationException e2) {
			l.fatal("Error occured while getting background sequences. Exiting.", e2);
			return;	
		}
		

		if (Boolean.parseBoolean(p.getProperty("gui"))) {
			startUpGUI(pList, sequenceManager, backgroundSequenceManager, pairs);
			return;
		}
		
		

		final File outputDir = new File(p.getProperty(Config.PropName.OUTPUT_DIR));
		outputDir.mkdirs();
		l.info("Output directory is: " + outputDir.getAbsolutePath());
		
		File gff = new File(outputDir, resultFileName);
		final BufferedWriter gffWriter;
		try {
			gffWriter = new BufferedWriter(new FileWriter(gff));
		} catch (IOException e1) {
			l.fatal(e1);
			return;
		}
		
		final File out = new File(outputDir, "out-" + new Date());
		out.mkdirs();
		Experiment exp;
		try {
			exp = new Experiment(pList, patternSet, pairs, sequenceManager, backgroundSequenceManager) {
				@Override
				public void process(AlignmentPrepared ap, List<Footprint> f) {
					File desc = new File(out, ap.getS1().getName() + "-" + ap.getS2().getName() + "-" + ap.getParameters().exportString());
					try {
						BufferedWriter bw = new BufferedWriter(new FileWriter(desc));
						for (Footprint fo : f) {
							bw.write(fo.exportString());
							bw.newLine();
							bw.write(fo.getS1Sites().toString());
							bw.newLine();
							bw.write(fo.getS2Sites().toString());
							bw.newLine();
							bw.newLine();
							if (fo.scorepv <= 0.05) {
								gffWriter.write(fo.gffString());
								gffWriter.newLine();
								gffWriter.flush();
							}
						}
						bw.close();
						logger.info("Output written to " + out.getAbsolutePath());
					} catch (IOException e) {
						logger.warn("Could not write result file", e);
					}
				}
			};
		} catch (ConfigurationException e) {
			l.error("Could not create the experiment.", e);
			return;
		}
		exp.perform();
		
		try {
			gffWriter.close();
		} catch (IOException e) {
			l.warn(e);
		}
	}

	private static BackgroundSequenceManager backgroundSetup(Threshold threshold, String patternSet) throws ConfigurationException {
		String path = p.getProperty(PropName.BACKGROUND_SEQUENCES_DIR);
		if (path == null) {
			throw new ConfigurationException(Config.PropName.SEQUENCES_DIR + " property should be set.");
		} else {
			l.info("Using " + path + " for backround sequences.");
			return new LocalBackgroundSequenceManager(path, threshold, Registry.getRegistry().getPatternSet(patternSet));
		}
	}

	private static SequenceManager sequenceSetup(List<SequencePair> pairs) throws ConfigurationException {
		String useDbString = p.getProperty(PropName.DOWNLOAD_SEQUENCES);
		if (useDbString != null && Boolean.parseBoolean(useDbString)) {
			l.info("Using Ensembl database for sequences");
			String downstreamString = p.getProperty(PropName.DOWNSTREAM_LENGTH);
			String upstreamString = p.getProperty(PropName.UPSTREAM_LENGTH);
			if (downstreamString == null || upstreamString == null) {
				throw new ConfigurationException(PropName.DOWNSTREAM_LENGTH + " and " + PropName.UPSTREAM_LENGTH + " should be both configured when using sequence download.");
			} else {
				int downstream = Integer.parseInt(downstreamString);
				int upstream = Integer.parseInt(upstreamString);
				return new DownloadingSequenceManager(downstream, upstream);
			}
		} else {
			l.info("Using locally stored sequences");
			String path = p.getProperty(Config.PropName.SEQUENCES_DIR); 
			if (path == null) {
				throw new ConfigurationException(Config.PropName.SEQUENCES_DIR + " property should be set.");
			} else {
				l.info("Using " + path + " for sequences.");
				return new LocalSequenceManager(path);
			}
		}
	}

	private static List<SequencePair> pairsSetup() {
		try {
			String pfp = p.getProperty(Config.PropName.PAIRS_FILE);
			List<SequencePair> pairs = Helpers.computePairs(new File(pfp));
			return pairs;
		} catch (IOException e1) {
			l.error("Something went wrong while getting the sequence pairs.",	e1);
			System.exit(1);
			return null;
		}
	}

	private static void propertiesSetup(String[] args) {
		try {
			ArgumentParser parser = new ArgumentParser(args);
			if (parser.getArgument(PropName.PROPERTIES_FILE) == null)
				p = Config.loadAppProperties();
			else {
				p = Config.loadAppPropertiesFromFile(parser.getArgument(PropName.PROPERTIES_FILE));
			}
			p.putAll(parser.getAll());
			p.putAll(System.getProperties());
		} catch (FileNotFoundException e1) {
			System.err.println("Property file not found. Exiting.");
			System.exit(1);
		} catch (IOException e1) {
			System.err
					.println("I/O error while retrieving application properties. Exiting.");
			System.exit(1);
		} catch (ArgumentParserException e) {
			System.err.println(e);
			System.exit(1);
		}
	}

	private static void startUpGUI(List<Parameters> ps, SequenceManager sequenceManager, BackgroundSequenceManager backgroundSequenceManager, List<SequencePair> pairs) {
		new app.gui.Billboard(ps, sequenceManager, backgroundSequenceManager, pairs).start();
	}

	private static void initialize() {
	}

	public static void getPatterns(String tfDir) {

		File pssmDir = new File(tfDir);
		File[] tfDirs = pssmDir.listFiles(new FileFilter() {
			public boolean accept(File f) {
				return f.isDirectory() && !f.getName().startsWith(".");
			}
		});
		
		for (File tfSetDir : tfDirs) {
			File[] pssmFiles = tfSetDir.listFiles(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					if (name.endsWith(".pfm"))
						return true;
					return false;
				}
			});
			
			final List<TFPattern> tfs = new LinkedList<TFPattern>();

			List<PSSM> pssms = new LinkedList<PSSM>();
			
			for (int i = 0; i < pssmFiles.length; i++) {
				try {
					PSSM pssm = Config.piof.buildSimple(
						new FileReader(pssmFiles[i]), pssmFiles[i].getName().split("\\.")[0]);
					pssms.add(pssm);
				} catch (IOException e) {
					l.warn("Could not create pssm from file " + pssmFiles[i].getName(), e);
				}
			}

			final OneTimeWorker<TFPattern> w = new OneTimeWorker<TFPattern>(2) {
				@Override
				public synchronized void afterExecute(TFPattern result) {
						tfs.add(result);
				}
			};
			for (final PSSM pssm : pssms) {
				try {
					w.submit(new Callable<TFPattern>() {
						public TFPattern call() throws Exception {
							return Config.tfpb.makeTFPattern(pssm, Background.getUniformBackground(pssm.getAlphabeth()));
						}
					});
				} catch (java.lang.NumberFormatException e) {
					l.warn("Could not parse pattern from file " + pssm.getName(), e);
				}
			}
			WorkerObject workerObject = w.execute();
			workerObject.run();
			try {
				workerObject.waitComplete();
			} catch (InterruptedException e1) {
				e1.printStackTrace();
				return;
			}
			
			Collections.sort(tfs, new Comparator<TFPattern>() {
				public int compare(TFPattern o1, TFPattern o2) {
					return o1.getName().compareTo(o2.getName());
				}
			});
			TFPattern[] patterns = tfs.toArray(new TFPattern[0]);
			PatternSet set = new PatternSet(tfSetDir.getName(), patterns);
			try {
				Registry.getRegistry().registerPatternSet(set);
			} catch (UserDataException e) {
				l.warn("Could not register pattern set with name "
						+ set.getName() + ". Set with this name already exists", e);
			}
		}
	}

	public static void setTestOutputStream() {
		System.setOut(new PrintStream(System.out) {
			@Override
			public void print(String s) {
				StackTraceElement e = new Exception().getStackTrace()[2];
				super.print(System.nanoTime() + " " + e.getClassName() + "."
						+ e.getMethodName() + "[" + e.getLineNumber() + "]: "
						+ s);
			}
		});
	}
}
