package logic.alignment;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import logic.multiset.MultiSet;
import logic.sequence.TFSite;
import logic.sequence.impl.PreparedSequence;
import logic.tfbs.impl.NoSuchLetter;
import logic.utils.ArrayView;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import app.intf.Parameters;

public class AlignmentPrepared {
	Log logger = LogFactory.getLog(this.getClass());
	private double[][] scoreMatrix;
	private Parameters p;
	private PreparedSequence s1;
	private PreparedSequence s2;
	private List<Footprint> f;
	private Set<PreparedSequence> bgSet;
	private ScoreMatrix sm;
	private double[][] scoreRanking;
	private double[][] sRanking;
	
	public AlignmentPrepared(PreparedSequence s1, PreparedSequence s2, Parameters p) {
		this.s1 = s1;
		this.s2 = s2;
		this.p = p;
	}
	
	public ScoreMatrix getScoreMatrix() {
		return sm;
	}
	
	public AlignmentPrepared(PreparedSequence s1, PreparedSequence s2, Parameters p, Set<PreparedSequence> bgPreps) {
		this(s1, s2, p);
		bgSet = bgPreps;
	}

	
	public AlignmentPrepared align() throws NoSuchLetter {
		MultiSet[] sets1 = getSets(s1.getPatterns());
		MultiSet[] sets2 = getSets(s2.getPatterns());
		
		if (bgSet != null) {
			computeRanking(sets1, bgSet);
		}
		// compute the highest scoring alignment
		sm = new ScoreMatrix(new SMatrix(new CutAndDiffMatrix(sets1, sets2), p.getAlpha(), p.getBeta(), p.getGamma()));
		scoreMatrix = sm.getResultMatrix();
		
		return this;
	}
	
	private MultiSet[] getSets(Collection<TFSite>[] patterns) {
		int seql = patterns.length;
		int matrixSize = (seql - p.getFrame())/p.getJump() + 1;
		MultiSet[] result = new MultiSet[matrixSize];
		for (int i = 0; i < matrixSize; i++) {
			result[i] = new MultiSet(
					new ArrayView<Collection<TFSite>>(patterns, (i)*p.getJump(), p.getFrame()));
		}
		return result;
	}

	private void computeRanking(MultiSet[] sets1, Set<PreparedSequence> bgSet) throws NoSuchLetter {
		double[][] sRanking = new double[sets1.length][bgSet.size()];
		double[][] scoreRanking = new double[sets1.length][bgSet.size()];
		int j = 0;
		for (PreparedSequence sequence : bgSet) {
			MultiSet[] sets = getSets(sequence.getPatterns());
			SMatrix sMatrix = new SMatrix(new CutAndDiffMatrix(sets1, sets), p.getAlpha(), p.getBeta(), p.getGamma());
			ScoreMatrix matrix = new ScoreMatrix(sMatrix);
			double[] sRowMax = sMatrix.getRowMax();		
			for (int i = 0; i < sRowMax.length; i++) {
				sRanking[i][j] = sRowMax[i];
			}
			double[] scoreRowMax = matrix.getRowMax();
			for (int i = 0; i < scoreRowMax.length; i++) {
				scoreRanking[i][j] = scoreRowMax[i];
			}
			j++;
		}
		for (int i = 0; i < sRanking.length; i++) {
			Arrays.sort(sRanking[i]);
		}
		for (int i = 0; i < scoreRanking.length; i++) {
			Arrays.sort(scoreRanking[i]);
		}
		
		this.sRanking = sRanking;
		this.scoreRanking = scoreRanking;
	}
	
	/*public List<Footprint> extractByPValue() {
		if (f == null) {
			double[][] workMatrix = deepCopy(scoreMatrix);
			f = new LinkedList<Footprint>();
			
		}
	}*/
	
	public List<Footprint> extractClustersDifferently() {
		if (f == null) {
			f = new LinkedList<Footprint>();
			double[][] scoreMatrix = sm.getResultMatrix();
			double[][] sMatrix = sm.getSMatrix();
			MaxIndex firstMax = findMax(scoreMatrix);
			for (int i = 1; i < scoreMatrix.length; i++) {
				int j = maxIndex(scoreMatrix[i]);
				if (scoreMatrix[i][j] > 0.01 * firstMax.max && sMatrix[i][j] > 0.) {
					int cur_i = i, cur_j = j;
					while (cur_i - 1 > 0 && cur_j - 1 > 0 && sMatrix[cur_i-1][cur_j-1] > 0.) {
						cur_i--;
						cur_j--;
					}
					//if (!(cur_i == i && cur_j == j)) {
					double sRank = getRank(cur_i, cur_j, i, j, sRanking);
					double scoreRank = getRank(cur_i, cur_j, i, j, scoreRanking);
					f.add(new Footprint(cur_i, cur_j, i, j, scoreMatrix[i][j], sRank, scoreRank, p, s1, s2));
			//		}
				}
			}
		}
		
		f = join(f);
		return f;
	}
	
	private List<Footprint> join(List<Footprint> f) {
		if (f.size() == 0)
			return f;
		Footprint[] footprints = f.toArray(new Footprint[0]);
		List<Footprint> result = new LinkedList<Footprint>();
		Arrays.sort(footprints, new Comparator<Footprint>() {
			public int compare(Footprint arg0, Footprint arg1) {
				return arg0.cell_i - arg1.cell_i == 0 ? (arg0.cell_j - arg1.cell_j) : arg0.cell_i - arg1.cell_i;
			}
		});
		Footprint prevF = footprints[0];
		for (int i = 1; i < footprints.length; i++) {
			Footprint footprint = footprints[i];
			if (footprint.cell_i == prevF.cell_i && footprint.cell_j == prevF.cell_j) {
				if (footprint.to_i > prevF.to_i) {
					prevF = footprints[i];
				}
			} else {
				result.add(prevF);
				prevF = footprints[i];
			}
		}
		result.add(prevF);
		Collections.sort(result, new Comparator<Footprint>() {
			public int compare(Footprint arg0, Footprint arg1) {
				return (int) (100 * arg0.scorepv - 100 * arg1.scorepv);
			}
		});
		return result;
	}

	private int maxIndex(double[] arr) {
		double curV = Double.NEGATIVE_INFINITY;
		int curI = -1;
		for (int i = 0; i < arr.length; i++) {
			if (arr[i] > curV) {
				curV = arr[i];
				curI = i;
			}
		}
		return curI;
	}
	
	private int myBinarySearch(double[] arr, double key) {
		return bSearch(arr, key, 0, arr.length);
	}

	private int bSearch(double[] arr, double key, int left, int right) {
		if (right - left == 1)
			return left;
		int c = (right + left) / 2;
		if (arr[c] < key)
			return bSearch(arr, key, c, right);
		else
			return bSearch(arr, key, left, c);
	}

	private double getRank(int cur_i, int cur_j, int max_i, int max_j, double[][] ranking) {
		double[][] s = sm.getSMatrix();
		int result = Integer.MIN_VALUE;
		for (int i = cur_i; i <= max_i; i++)
			result = Math.max(result, myBinarySearch(ranking[i], s[i][i - cur_i + cur_j]));
		return 1 - ((result - 0.0)/ranking[0].length);
	}

	private class MaxIndex {
		private int i,j;
		private double max;
		private MaxIndex(int i, int j, double max) {
			this.i = i; this.j = j; this.max = max;
		}
	}
	
	private MaxIndex findMax(double[][] workMatrix) {
		double result = Double.NEGATIVE_INFINITY;
		int mi = 0, mj = 0;
		for (int i = 0; i < workMatrix.length; i++)
			for (int j = 0; j < workMatrix[i].length; j++) {
				if (workMatrix[i][j] > result) {
					result = workMatrix[i][j];
					mi = i;
					mj = j;
				}
			}
		return new MaxIndex(mi, mj, result);
	}

	public static AlignmentPrepared alignPrepared(PreparedSequence s1, PreparedSequence s2, Parameters p) throws NoSuchLetter {
		return new AlignmentPrepared(s1, s2, p).align();
	}
	
	public Parameters getParameters() {
		return p;
	}
	
	public static class Footprint {
		public final int cell_i, cell_j, to_i, to_j;
		public final double score;
		public final PreparedSequence ps1;
		public final PreparedSequence ps2;
		public final Parameters p;
		public final double spv, scorepv;
		
		private Footprint(int cell_i, int cell_j, int to_i, int to_j, double score, double sRank, double scoreRank, Parameters p2, PreparedSequence ps1, PreparedSequence ps2) {
			this.cell_i = cell_i;
			this.cell_j = cell_j;
			this.to_i = to_i;
			this.to_j = to_j;
			this.score = score;
			this.spv = sRank;
			this.scorepv = scoreRank;
			this.ps1 = ps1;
			this.ps2 = ps2;
			this.p = p2;
		}
		
		private int getStartS1() {
			return cell_i*p.getJump();
		}
		private int getStartS2() {
			return cell_j*p.getJump();
		}
		private int getEndS1() {
			return (to_i)*p.getJump() - 1 + p.getFrame();
		}
		private int getEndS2() {
			return (to_j)*p.getJump() -1 + p.getFrame();
		}
		
		public String toString() {
			return exportString();
		}

		public Collection<TFSite> getS1Sites() {
			return ps1.getSites(getStartS1(), getEndS1() - getStartS1());
		}
		public Collection<TFSite> getS2Sites() {
			return ps2.getSites(getStartS2(), getEndS2() - getStartS2());
		}
		
		
		/**
		 * Export string. Export format is:
		 * (first sequence name) (second sequence name) (alignment score) (p-value) (first index on s1 (from 0)) (last index on first) (first index on s2) (last index on s2) 
		 * @return export string
		 */
		public String exportString() {
			StringBuffer sb = new StringBuffer();
			sb.append(ps1.getName()); sb.append("\t");
			sb.append(ps2.getName()); sb.append("\t");
			sb.append(Math.round(score*100)/100.0); sb.append("\t");
			sb.append(Math.round(100*spv)/100.0); sb.append("\t");
			sb.append(Math.round(100*scorepv)/100.0); sb.append("\t");
			sb.append(getStartS1()); sb.append("\t");
			sb.append(getEndS1()); sb.append("\t");
			sb.append(getStartS2()); sb.append("\t");
			sb.append(getEndS2()); sb.append("\t");
			return sb.toString();
		}
		
		public String gffString() {
			StringBuffer sb = new StringBuffer();
			sb.append(ps1.getName()); sb.append("\t");
			sb.append("billboard_prediction"); sb.append("\t");
			sb.append("enhancer"); sb.append("\t");
			sb.append(getStartS1()); sb.append("\t");
			sb.append(getEndS1()); sb.append("\t");
			sb.append(Math.round(100*scorepv)/100.0); sb.append("\t");
			sb.append("."); sb.append("\t");
			sb.append("."); sb.append("\t");
			sb.append("."); sb.append("\t"); 
			sb.append("Against " + ps2.getName());
			return sb.toString();
		}
	}

	public void dumpScoreMatrix(OutputStream os) throws IOException {
		dumpMatrix(os, scoreMatrix);

	}
	
	private void dumpMatrix(OutputStream os, double[][] matrix) throws IOException {
		OutputStreamWriter osw = new OutputStreamWriter(os);
		for (int i = 1; i < matrix.length; i++) {
			for (int j = 1; j < matrix[0].length; j++) {
				if (j > 1)
					osw.write("\t");
				osw.write("" + matrix[i][j]);
			}
			osw.write("\n");
		}
		osw.flush();
		
	}
	
	public String toString() {
		return s1.getName() + "-" + s2.getName() + ":" + p;
	}

	public PreparedSequence getS1() {
		return s1;
	}

	public PreparedSequence getS2() {
		return s2;
	}

}
