package logic.sequence.impl;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import logic.sequence.Alphabet;
import logic.sequence.DNASequence;
import logic.sequence.SequenceBuilder;
import logic.sequence.TFSite;
import logic.tfbs.PatternSet;
import logic.tfbs.TFPattern;

public final class DNASequenceImpl implements DNASequence {
	
	private String name;
	private char[] sequence;

	public final char getAt(int i) {
		return sequence[i];
	}

	public String getName() {
		return name;
	}

	public int length() {
		return sequence.length;
	}
	
	private DNASequenceImpl(String name, String sequence) throws BadAlphabet {
		if (name == null || sequence == null)
			throw new NullPointerException();
		this.name = name;
		this.sequence = sequence.toLowerCase().toCharArray();
	}

	
	private static class SequenceBuilderImpl implements SequenceBuilder {

		public DNASequence createDNASequence(String name, String sequence, boolean acceptAmbigous) throws BadAlphabet {
			if (acceptAmbigous) {
				if (!checkSequence(sequence.toLowerCase(), Alphabet.AMBIGOUS_ALPHABET)) {
					throw new BadAlphabet();
				}
			} else {
				if (!checkSequence(sequence.toLowerCase(), Alphabet.UNAMBIGOUS_ALPHABET)) {
					throw new BadAlphabet();
				}
			}
			return new DNASequenceImpl(name, sequence);
		}
		
		private static boolean checkSequence(String sequence, String alphabet) {
			return sequence.matches("[" + alphabet + "]+");
		}

		public static SequenceBuilder getInstance() {
			return new SequenceBuilderImpl();
		}
		
	}
	
	public static SequenceBuilder getBuilder() {
		return SequenceBuilderImpl.getInstance();
	}

/*	public Collection<TFSite>[] matchPatterns(Collection<TFPattern> patterns, double threshold) throws NoSuchLetter {
		return matchPatterns(patterns.toArray(new TFPattern[0]), threshold);
	}*/
	
	public String toString() {
		return ">" + this.name + "\n" + String.valueOf(this.sequence);
	}

	@SuppressWarnings("unchecked")
	public Collection<TFSite>[] matchPatterns(PatternSet patternSet) {
		List[] collections = new LinkedList[this.length()];
		for (int i = 0; i < collections.length; i++) {
			collections[i] = new LinkedList<TFPattern>();
		}
		List<TFSite>[] result = collections;
		for (TFPattern pattern : patternSet) {
			double[] ds = pattern.matchAll(this);
			double[] dsc = pattern.matchAllComplimentary(this);
			for (int i = 0; i < ds.length; i++) {
				if (ds[i] >= pattern.getBalancedThreshold())
					result[i].add(new TFSiteImpl(pattern, ds[i], i, true));
				if (dsc[i] >= pattern.getBalancedThreshold())
					result[i].add(new TFSiteImpl(pattern, dsc[i], i, false));
			}
		}
		return result;
	}
	
	@SuppressWarnings("unchecked")
	public Collection<TFSite>[] matchPatterns(PatternSet patternSet, double threshold) {
		List[] collections = new LinkedList[this.length()];
		for (int i = 0; i < collections.length; i++) {
			collections[i] = new LinkedList<TFPattern>();
		}
		List<TFSite>[] result = collections;
		for (TFPattern pattern : patternSet) {
			double[] ds = pattern.matchAll(this);
			double[] dsc = pattern.matchAllComplimentary(this);
			for (int i = 0; i < ds.length; i++) {
				if (ds[i] >= threshold)
					result[i].add(new TFSiteImpl(pattern, ds[i], i, true));
				if (dsc[i] >= threshold)
					result[i].add(new TFSiteImpl(pattern, dsc[i], i, false));
			}
		}
		return result;
	}

	public String getFragment(int offset, int length) {
		char[] result = new char[length];
		for (int i = 0; i < length; i++) {
			result[i] = sequence[offset+i];
		}
		return String.valueOf(result);
	}
	
	public void exportFastaStream(OutputStream os) throws IOException {
		BufferedWriter bw = new BufferedWriter(
				new OutputStreamWriter(os));
		exportFastaStream(bw);
	}

	public void exportFastaStream(BufferedWriter bw) throws IOException {
		bw.write(">");
		bw.write(this.name);
		bw.newLine();
		bw.write(this.sequence);
		bw.newLine();
		bw.flush();
	}



}
