package logic.sequence.io;

import java.io.IOException;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import logic.sequence.DNASequence;
import logic.sequence.impl.DNASequenceImpl;
import logic.sequence.io.db.EnsemblConnection;
import logic.sequence.io.db.QueryException;
import logic.sequence.io.db.SqlLib;
import logic.sequence.io.db.sql.entity.AssemblyRegion;
import logic.sequence.io.db.sql.entity.Gene;
import logic.sequence.io.db.sql.entity.Transcript;
import logic.sequence.io.db.sql.entity.manager.StringManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import app.ensembl.IdNotRecognizedException;
import app.ensembl.Species;
import app.exception.BillboardException;

public class EnsemblDBReader {
	
	Log log = LogFactory.getLog(EnsemblDBReader.class);
	
	private EnsemblConnection c;
	private static EnsemblDBReader edbr;
	
	private EnsemblDBReader() throws BillboardException {
		try {
			c = EnsemblConnection.getConnection();
		} catch (Exception e) {
			throw new BillboardException(e);
		}
	}
	
	public static EnsemblDBReader getReader() throws BillboardException {
		if (edbr == null)
			edbr = new EnsemblDBReader();
		return edbr;
	}
	
	public String getGeneStableIdByTranscriptStableId(String stableId) throws SQLException, QueryException, IdNotRecognizedException {
		Species s = Species.getForEnsembleId(stableId);
		String geneStableId = (String) c.uniqueResult(SqlLib.getGeneStableIdForTranscriptStableId(stableId, s));
		return geneStableId;
	}
	
	public List<Gene> getGeneIdentifiers(Species s) throws SQLException {
		return (List<Gene>) c.list(SqlLib.getGenesForSpecies(s), Gene.getEntityManager());
	}
	
	public DNASequence getByStableId(String id, int beforeStart, int afterStart, boolean allowAmbigous) throws SQLException, QueryException, IOException, BillboardException {
		Species s = Species.getForEnsembleId(id);
		if (Species.isGene(id)) {
			Gene t = (Gene) c.uniqueResult(SqlLib.getGeneForStableId(id), Gene.getEntityManager());
			int start, end;
			if (t.strand) {
				start = t.start- beforeStart;
				end = t.start + afterStart;
			} else {
				start = t.end - afterStart;
				end = t.end + beforeStart;
			}
			return DNASequenceImpl.getBuilder().createDNASequence(id, getSequence(t.seqRegionId,
					s, start, end, t.strand), allowAmbigous);
		} else {
			Transcript t = (Transcript) c.uniqueResult(SqlLib.getTranscriptForStableId(id), Transcript.getEntityManager());
			int start, end;
			if (t.seqRegionStrand) {
				start = t.seqRegionStart - beforeStart;
				end = t.seqRegionStart + afterStart;
			} else {
				start = t.seqRegionEnd - afterStart;
				end = t.seqRegionEnd + beforeStart;
			}	
			return DNASequenceImpl.getBuilder().createDNASequence(id, getSequence(t.seqRegionId,
					s, start, end, t.seqRegionStrand), allowAmbigous);
		}
	}
	
	public String getSequence(int chromosomeSegRegion, Species s,
			int first, int last, boolean strand) throws SQLException,
			IOException, BillboardException {
		String str = SqlLib.getAssemblyFragments(chromosomeSegRegion, first,
				last, s);
		List<AssemblyRegion> assemblies = (List<AssemblyRegion>) c.list(str, AssemblyRegion.getEntityManager());
		Collections.sort(assemblies, new Comparator<AssemblyRegion>() {
			public int compare(AssemblyRegion a, AssemblyRegion b) {
				return a.asmStart - a.asmEnd;
			}
		});

		Iterator<AssemblyRegion> it = assemblies.iterator();
		AssemblyRegion ar = null;
		int nextToRead = first;
		StringBuffer sb = new StringBuffer();
		do {
			ar = it.next();
			int start, end;
			if (ar.asmStart > nextToRead) {
				int times = ar.asmStart - nextToRead;
				sb.append(n(times));
				nextToRead = ar.asmStart;
			}
			start = nextToRead - ar.asmStart + ar.cmpStart;
			if (ar.asmEnd > last) {
				end = last - ar.asmStart + ar.cmpStart;
			} else {
				end = ar.cmpEnd;
			}
			nextToRead += (end - start + 1);
			String sql = SqlLib.getSequence(ar.cmpId, start, end, s);
			List<String> seq = (List<String>) c.list(sql, new StringManager());
			sb.append(seq.get(0));
		} while (it.hasNext() && nextToRead <= last);
		if (nextToRead <= last) {
			sb.append(n(last - nextToRead));
		}
		String sequence = strand ? sb.toString() : reverseComplement(sb
				.toString());

		return sequence;
	}
	
	private static String n(int times) {
		char[] ns = new char[times];
		Arrays.fill(ns, 'n');
		return new String(ns);
	}
	
	private static String reverseComplement(String seq) {
		StringBuilder sb = new StringBuilder();
		for (int i = seq.length() - 1; i >= 0; i--) {
			char r = reverse(seq.charAt(i));
			sb.append(r);
		}
		return sb.toString();
	}

	private static final char reverse(char r) {
		switch (r) {
		case 'a':
		case 'A':
			return 'T';
		case 't':
		case 'T':
			return 'A';
		case 'g':
		case 'G':
			return 'C';
		case 'c':
		case 'C':
			return 'G';
		default:
			return r;
		}
	}

}
