package logic.tfbs.impl;

import logic.sequence.DNASequence;
import logic.tfbs.PSSM;
import logic.tfbs.TFPattern;
import app.exception.BillboardError;

public class TFPatternImpl implements TFPattern {
	
	private PSSM pssm;
	protected double[][] logOdds = null;
	protected double[][] logOddsComplimentary = null;
	protected double[][] pwm;
	protected double[][] pspm;
	private int id;
	private Background b = null;
	private ScoreDistribution distr;
	private double fprThreshold;
	private double fnrThreshold;
	private double balancedThreshold;
	
	private static volatile int idCounter = 0;

	public int getId() {
		return id;
	}
	
	TFPatternImpl(PSSM pssm) throws NoSuchLetter {
		this.pssm = pssm;
		this.pspm = computePSPMMatrix(pssm);
		this.pwm = computeLogMatrix(pssm);
		this.id = idCounter++;
	}
	
	TFPatternImpl(PSSM pssm, Background background) throws NoSuchLetter {
		this(pssm);
		if (background != null)
			setBackground(background);
	}

	private void setBackground(Background background) throws NoSuchLetter {
		this.b = background;
		recomputeLogOddsMatrix();
		// TODO: add configurability of this feature
		distr = new ScoreDistribution(1000);
		fprThreshold = distr.countFPRThreshold(0.001);
		fnrThreshold = distr.countFNRThreshold(0.1);
		balancedThreshold = distr.countBalancedThreshold(1000);
	}

	private void recomputeLogOddsMatrix() throws NoSuchLetter {
		this.logOdds = new double[pssm.getAlphabethSize()][pssm.getMatrixWidth()];
		this.logOddsComplimentary = new double[pssm.getAlphabethSize()][pssm.getMatrixWidth()];
		for (int i = 0; i < pssm.getAlphabethSize(); i++) {
			double uniFactor = Math.log(b.getForLetter(pssm.getAlphabeth()[i]))/Math.log(2);
			for (int j = 0; j < pssm.getMatrixWidth(); j++) {
				logOdds[i][j] = pwm[i][j] - uniFactor;
				logOddsComplimentary[i][j] =
					pwm[pssm.getComplimentaryRow(pssm.getAlphabeth()[i])][pssm.getMatrixWidth() - 1 - j] - uniFactor;
			}
		}
	}

	private double[][] computePSPMMatrix(PSSM pssm) throws NoSuchLetter {
		double[][] pspm = new double[pssm.getAlphabethSize()][pssm.getMatrixWidth()];
		int sigma = pssm.getAlphabethSize();
		for (int i = 0; i < pssm.getAlphabethSize(); i++)
			for (int j = 0; j < pssm.getMatrixWidth(); j++) {
				double denominator = sigma + pssm.getQuantity(j);
				pspm[pssm.getRow(pssm.getAlphabeth()[i])][j] = (pssm.getValueAt(i, j) + 1.0) / denominator;
			}
		return pspm;
	}

	private double[][] computeLogMatrix(PSSM pssm) throws NoSuchLetter {
		double[][] logM = new double[pssm.getAlphabethSize()][pssm.getMatrixWidth()];
		for (int i = 0; i < pssm.getAlphabethSize(); i++)
			for (int j = 0; j < pssm.getMatrixWidth(); j++) {
				logM[i][j] =
					Math.log(pspm[i][j])/Math.log(2);
			}
		return logM;
	}

	public double[] matchAll(DNASequence sequence) {
		return doMatchAll(sequence, logOdds);
	}
	
	private double[] doMatchAll(DNASequence sequence, double[][] logMatrix) {
		double[] result =
			new double[sequence.length() - pssm.getMatrixWidth() + 1];
		for (int i = 0; i < result.length; i++) {
			try {
				for (int j = 0; j < pssm.getMatrixWidth(); j++) {
					result[i] += logMatrix[pssm.getRow(sequence.getAt(i + j))][j];
				}
			} catch (NoSuchLetter e) {
				result[i] = Double.NaN;
			}
		}
		return result;
	}
	
	public double[] matchAllComplimentary(DNASequence sequence) {
		return doMatchAll(sequence, logOddsComplimentary);
	}
	
	
	
	private class ScoreDistribution {
		private int precision;
		private double min, max;
		private double minValue;
		private double maxValue;
		private double patternMinValue;
		private double patternMaxValue;
		private double step;
		private double[] backgroundDensity;
		private double[] patternDensity;
		
		private ScoreDistribution(int precision) {
			this.precision = precision * TFPatternImpl.this.length();
			this.minValue = getMinValue();
			this.maxValue = getMaxValue();
			this.patternMinValue = getPatternMinValue();
			this.patternMaxValue = getPatternMaxValue();
			this.min = Math.min(minValue, patternMinValue);
			this.max = Math.max(maxValue, patternMaxValue);
			this.step = (this.max - this.min) / this.precision;
			computeBackgroundDensity();
			computePatternDensity();
		}
		
		private int add(int i) {
			return Math.max(0, Math.min(i, precision - 1));
		}
		
		private void computeBackgroundDensity() {
			this.backgroundDensity = new double[precision];
			this.backgroundDensity[-mapScoreZeroRelative(min)] = 1;
			try {
				for (int j = 0; j < length(); j++) {
					double[] colDensity = new double[precision];
					for (int i = 0; i < logOdds.length; i++) {
						int index = mapScoreZeroRelative(logOdds[i][j]);
						for (int di = 0; di < precision; di++)
							colDensity[add(index + di)] += this.backgroundDensity[di] * b.getForLetter(pssm.getAlphabeth()[i]); 
					}
					this.backgroundDensity = colDensity;
				}
			} catch (NoSuchLetter e) {
				throw new BillboardError(e);
			}
		}
		
		private void computePatternDensity() {
			this.patternDensity = new double[precision];
			this.patternDensity[-mapScoreZeroRelative(min)] = 1;
			for (int j = 0; j < length(); j++) {
				double[] colDensity = new double[precision];
				for (int i = 0; i < logOdds.length; i++) {
					int index = mapScoreZeroRelative(logOdds[i][j]);
					for (int di = 0; di < precision; di++)
						colDensity[add(index + di)] += this.patternDensity[di] * pspm[i][j]; 
				}
				this.patternDensity = colDensity;
			}
		}
		
/*		private int mapScore(double score) {
			return (int) ((score - min) / step);
		}*/
		
		private final int mapScoreZeroRelative(double score) {
			
			return (int) (Math.signum(score) * (Math.abs(score) / step));
		}
		
		public double countFPRThreshold(double fpr) {
			double pr = 0;
			for (int i = precision - 1; i > 0; i--) {
				pr += backgroundDensity[i];
				if (pr >= fpr)
					return min + step * i;
			}
			throw new BillboardError("Could not count fpr threshold for pattern " + getName());
		}

		public double countFNRThreshold(double fnr) {
			double pr = 0;
			for (int i = 0; i < precision; i++) {
				pr += patternDensity[i];
				if (pr >= fnr)
					return min + step * i;
			}
			throw new BillboardError("Could not count fnr threshold for pattern " + getName());

		}
		
		public double countBalancedThreshold(int proportion) {
			int i = precision;
			double fpr = 0, fnr = 1;
			while (fpr * proportion < fnr) {
				i--;
				fpr += backgroundDensity[i];
				fnr -= patternDensity[i];
			}
			return min + i * step;
		}

		private double getMinValue() {
			double result = Double.POSITIVE_INFINITY;
			for (int i = 0; i < logOdds.length; i++)
				for (int j = 0; j < logOdds[0].length; j++) {
					if (logOdds[i][j] < result)
						result = logOdds[i][j];
				}
			return result;
		}

		private double getMaxValue() {
			double result = Double.NEGATIVE_INFINITY;
			for (int i = 0; i < logOdds.length; i++)
				for (int j = 0; j < length(); j++) {
					if (logOdds[i][j] > result)
						result = logOdds[i][j];
				}
			return result;
		}
		
		private double getPatternMinValue() {
			double result = 0;
			double columnMin = Double.POSITIVE_INFINITY;
			for (int j = 0; j < length(); j++) {
				for (int i = 0; i < logOdds.length; i++)
					if (logOdds[i][j] < columnMin)
						columnMin = logOdds[i][j];
				result += columnMin;
				columnMin = Double.POSITIVE_INFINITY;
			}
			return result;
		}
		
		private double getPatternMaxValue() {
			double result = 0;
			double columnMax = Double.NEGATIVE_INFINITY;
			for (int j = 0; j < length(); j++) {
				for (int i = 0; i < logOdds.length; i++)
					if (logOdds[i][j] > columnMax)
						columnMax = logOdds[i][j];
				result += columnMax;
				columnMax = Double.NEGATIVE_INFINITY;
			}
			return result;
		}
	}

	public int length() {
		return pssm.getMatrixWidth();
	}

	public String getName() {
		return pssm.getName();
	}
	
	public String toString() {
		return pssm.getName();
	}
	
	public PSSM getPSSM() {
		return this.pssm;
	}

	public double[][] getLogOddsMatrix() {
		return logOdds;
	}

	public double[][] getPSPM() {
		return pspm;
	}

	public double getFnrThreshold() {
		return fnrThreshold;
	}

	public double getFprThreshold() {
		return fprThreshold;
	}

	public double getBalancedThreshold() {
		return balancedThreshold;
	}
}
