package logic.threads;

import java.util.LinkedList;
import java.util.Queue;
import java.util.Vector;
import java.util.concurrent.Callable;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class OneTimeWorker<V> {
	
	private static Log l = LogFactory.getLog(OneTimeWorker.class);

	private int nThreads;
	private LinkedList<Callable<V>> jobs = new LinkedList<Callable<V>>();

	public OneTimeWorker(int nThreads) {
		this.nThreads = nThreads;
	}
	
	public void submit(Callable<V> c) {
		jobs.add(c);
	}
	
	public void afterExecute(V v) {};
	public void onException(Exception e) {};
	
	public WorkerObject execute() {
		Queue<Runnable> q = new LinkedList<Runnable>();

		for (int i = 0; i < jobs.size(); i++) {
			final Callable<V> c = jobs.get(i);
			q.offer(new Runnable() {
				public void run() {
					try {
						V v = c.call();
						afterExecute(v);
					} catch (Exception e) {
						onException(e);
					}
				}
			});

		}
		return new WorkerObject(nThreads, q);
	}
	
	public static interface WorkerObjectMBean {
		public int getJobsLeft();
		public int getThreadNo();
		public void increaseThreadNo();
		public void decreaseThreadNo();
	}
	
	public static class WorkerObject implements WorkerObjectMBean {
		private Queue<Runnable> jobs;
		private Vector<PoolWorker> workers;
		private int all;
		
		public WorkerObject(int nThreads, Queue<Runnable> jobs) {
			this.jobs = jobs;
			all = jobs.size();
			workers = new Vector<PoolWorker>(nThreads);
			for (int i = 0; i < nThreads; i++) {
				PoolWorker poolWorker = new PoolWorker();
				poolWorker.setName("PoolWorker-" + i);
				workers.add(poolWorker); 
			}
		}
		
		public int getJobsLeft() {
			synchronized(jobs) {
				return jobs.size();
			}
		}
		
		public class PoolWorker extends Thread {
			private boolean stop = false;
			public void run() {
				final long cur = System.currentTimeMillis();
				int i = 0;
				Runnable r;
				while (true) {
					if (stop)
						return;
					synchronized(jobs) {
						r = jobs.poll();
						if (r == null)
							return;
					}
					try {
						r.run();
					} catch (RuntimeException e) {
						
					} finally {
						i++;
					}
					long upto = System.currentTimeMillis() - cur;
					l.info(jobs.size() + "/" + all + ". [Est: " + upto
							* jobs.size() / (all - jobs.size()) / 1000 / 60
							+ " minutes]");
				}
			}
			public void setStop() {
				stop = true;
			}
		}
		
		public void run() {
			for (PoolWorker pw : workers)
				pw.start();
		}
		
		public void waitComplete() throws InterruptedException {
			for (PoolWorker pw : workers) {
				pw.join();
			}
		}
		
		public void interrupt() {
			synchronized(jobs) {
				jobs.clear();
			}
		}

		public int getThreadNo() {
			synchronized(workers) {
				return workers.size();
			}
		}

		public void increaseThreadNo() {
			PoolWorker pw = new PoolWorker();
			pw.setName("PoolWorker");
			pw.start();
			synchronized(workers) {
				workers.add(pw);
			}
		}

		public void decreaseThreadNo() {
			PoolWorker pw;
			synchronized(workers) {
				pw = workers.lastElement();
				workers.remove(pw);
			}
			pw.setStop();
		}
	}
}
