package logic.threads;

import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;


public class ThreadPool {
	
	private Logger l;

	private final PoolWorker[] pool;
	private final LinkedList<Runnable> queue;
	private final IdleMonitor im = new IdleMonitor();

	public ThreadPool(int threads, String name) {
		queue = new LinkedList<Runnable>();
		pool = new PoolWorker[threads];
		l = Logger.getLogger(name);
		im.start();
		for (int i = 0; i < threads; i++) {
			pool[i] = new PoolWorker();
			pool[i].start();
		}
	}
	
	public void execute(Runnable r) {
		synchronized(queue) {
			queue.addLast(r);
			queue.notify();
		}
	}
	
	private class PoolWorker extends Thread {
		public void run() {
			im.busy();
			Runnable r;
			while (true) {
				synchronized(queue) {
					if (queue.isEmpty()) {
						try {
							im.idle();
							queue.wait();
							im.busy();
						} catch (InterruptedException ignore) { return; }
					}
					r = queue.removeFirst();
				}
				try {
					r.run();
				} catch (RuntimeException e) {
					l.log(Level.WARNING, "", e);
				}
			}
		}
	}
	
	private class IdleMonitor extends Thread {
		private int busy = 0;
		private Object lock = new Object();
		
		public IdleMonitor () {
			super();
		}
		public void run() {
			synchronized(lock) {
				while (true) {
					if (busy == 0) {
						synchronized(this) {
							notifyAll();
						}
					}
					try {
						lock.wait();
					} catch (InterruptedException ignore) {}
				}
			}
		}
		public void busy() {
			synchronized(lock) {
				busy++;
			}
		}
		public void idle() {
			synchronized(lock) {
				busy--;
				lock.notify();
			}
		}
	}

	public void waitIdle() throws InterruptedException {
		synchronized(im) {
			im.wait();
		}
	}
	
	public void destroyThreads() {
		for (Thread t : pool) {
			t.interrupt();
		}
	}

	public void interrupt() {
		synchronized(queue) {
			queue.clear();
			System.out.println("job queue cleared");
		}
	}
}
